package utils

import (
	"bufio"
	"errors"
	"log"
	"os"
	"strconv"
	"strings"
)

var (
	errEmptyValue   = errors.New("empty value")
	errNoSuchOption = errors.New("no such option")
)

func LookForRawOptionInConfig(configFileName string, optionName string) (string, error) {
	c, err := os.Open(configFileName)
	if err != nil {
		return "", err
	}
	defer c.Close()

	scan := bufio.NewScanner(c)
	for scan.Scan() {
		if !IsCommented(scan.Text(), "[") && !IsCommented(scan.Text(), "/") {
			param := strings.Split(scan.Text(), `=`)
			if param[0] == optionName {
				if param[1] != "" {
					return param[1], nil
				} else {
					return "", errEmptyValue
				}
			}
		}
	}
	return "", errNoSuchOption
}

func LookForStringOptionInConfig(configFileName string, optionName string, defaultValue string) string {
	o, err := LookForRawOptionInConfig(configFileName, optionName)
	if err != nil {
		log.Printf("Can't set %s: %s; using default value: %s\n", optionName, err.Error(), defaultValue)
		return defaultValue
	}
	log.Printf("Set %s: %s\n", optionName, o)
	return o
}

func LookForBoolOptionInConfig(configFileName string, optionName string, defaultValue bool) bool {
	o, err := LookForRawOptionInConfig(configFileName, optionName)
	if err != nil {
		log.Printf("Can't set %s: %s; using default value: %t\n", optionName, err.Error(), defaultValue)
		return defaultValue
	}
	v, err := strconv.ParseBool(o)
	if err != nil {
		log.Printf("Can't set %s: %s; using default value: %t\n", optionName, err.Error(), defaultValue)
		return defaultValue
	}
	log.Printf("Set %s: %t\n", optionName, v)
	return v
}

func LookForIntOptionInConfig(configFileName string, optionName string, defaultValue int) int {
	o, err := LookForRawOptionInConfig(configFileName, optionName)
	if err != nil {
		log.Printf("Can't set %s: %s; using default value: %d\n", optionName, err.Error(), defaultValue)
		return defaultValue
	}
	v, err := strconv.Atoi(o)
	if err != nil {
		log.Printf("Can't set %s: %s; using default value: %d\n", optionName, err.Error(), defaultValue)
		return defaultValue
	}
	log.Printf("Set %s: %d\n", optionName, v)
	return v
}

func IsCommented(line string, symbol string) bool {
	if line != "" {
		return string(line[0]) == symbol
	}
	return true
}
